/*
 * Copyright (C) 2013 Samsung Electronics
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/io.h>
#include <netdev.h>
#include <asm/arch/cpu.h>
#include <asm/arch/clock.h>
#include <asm/arch/power.h>
#include <asm/arch/gpio.h>
#include <asm/arch/mmc.h>
#include <asm/arch/pinmux.h>
#include <asm/arch/sromc.h>
#include <asm/arch/sysreg.h>
#include "pmic.h"
#include "setup.h"
#include "util.h"

#define rGPX1CON *(volatile uint32_t *)(0x13400C20)
#define rGPX1DAT *(volatile uint32_t *)(0x13400C24)
#define rGPX1PUD *(volatile uint32_t *)(0x13400C28)

#define rGPG0CON	*(volatile unsigned long *)(0x14000080)
#define rGPG0DAT	*(volatile unsigned long *)(0x14000084)
#define rGPG0PUD	*(volatile unsigned long *)(0x14000088)

#define rGPC0CON	*(volatile unsigned long *)(0x13410000)
#define rGPC0DAT	*(volatile unsigned long *)(0x13410004)

#define rGPD1CON	*(volatile unsigned long *)(0x134100A0)
#define rGPD1DAT	*(volatile unsigned long *)(0x134100A4)
#define rGPD1PUD	*(volatile unsigned long *)(0x134100A8)

DECLARE_GLOBAL_DATA_PTR;
unsigned int pmic;
unsigned int nr_dram_banks = 8;

typedef struct {
	uint8_t pin;
	uint8_t cfg;
	uint8_t pud;
	uint8_t drv;
	uint8_t rate;
	uint8_t pdn_con;
	uint8_t pdn_pull;
} gpio_pin_cfg_t;

typedef struct {
	struct s5p_gpio_bank *bank;
	gpio_pin_cfg_t pin_cfgs[8];
} gpio_bank_list_t;

#define GPIO_DEFAULT	((uint8_t)-1U)	// indicates that we don't want to change the default setting
#define TERMINATE_PINLIST {(uint8_t)-1U, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT}

static void setup_pins(gpio_bank_list_t *list, int num);
static void setup_default_pins(void);

static int init_nr_dram_banks(void)
{
	nr_dram_banks = 8;
}

static void display_bl1_version(void)
{
	char bl1_version[9] = {0};

	/* display BL1 version */
#if defined(CONFIG_TRUSTZONE_ENABLE)
	printf("\nTrustZone Enabled BSP");
	strncpy(&bl1_version[0], (char *)(CONFIG_PHY_IRAM_NS_BASE + 0x810), 8);
	printf("\nBL1 version: %s\n", &bl1_version[0]);
#endif
}

static void display_pmic_info(void)
{
	board_id_t id = get_board_info();

	uint8_t vdd_kfc, vdd_arm, vdd_int, vdd_mif, vdd_g3d;

	if(id.is_dev_board || id.revision == 0x7) {
		/*
		 * Dev Board - PMIC is on IIC0
		 * (Proto FF has the same setup)
		 */

		IIC0_ESetport();
		IIC0_ERead(S2MPS11_ADDR, BUCK6_CTRL2, &vdd_kfc);
		IIC0_ERead(S2MPS11_ADDR, BUCK2_CTRL2, &vdd_arm);
		IIC0_ERead(S2MPS11_ADDR, BUCK3_CTRL2, &vdd_int);
		IIC0_ERead(S2MPS11_ADDR, BUCK1_CTRL2, &vdd_mif);
		IIC0_ERead(S2MPS11_ADDR, BUCK4_CTRL2, &vdd_g3d);

	} else {
		/* FF Board later than Proto has PMIC on IIC4 - GPA2{0:1} */
		IIC4_ESetport();
		IIC4_ERead(S2MPS11_ADDR, BUCK6_CTRL2, &vdd_kfc);
		IIC4_ERead(S2MPS11_ADDR, BUCK2_CTRL2, &vdd_arm);
		IIC4_ERead(S2MPS11_ADDR, BUCK3_CTRL2, &vdd_int);
		IIC4_ERead(S2MPS11_ADDR, BUCK1_CTRL2, &vdd_mif);
		IIC4_ERead(S2MPS11_ADDR, BUCK4_CTRL2, &vdd_g3d);
	}

	printf("VDD_KFC: 0x%x\n", (uint32_t)vdd_kfc);
	printf("VDD_ARM: 0x%x\n", (uint32_t)vdd_arm);
	printf("VDD_INT: 0x%x\n", (uint32_t)vdd_int);
	printf("VDD_MIF: 0x%x\n", (uint32_t)vdd_mif);
	printf("VDD_G3D: 0x%x\n", (uint32_t)vdd_g3d);
}

static void display_boot_device_info(void)
{
	struct exynos5_power *pmu = (struct exynos5_power *)EXYNOS5_POWER_BASE;
	int OmPin;

	OmPin = readl(&pmu->inform3);

	printf("\nChecking Boot Mode ...");

	if (OmPin == BOOT_MMCSD)
		printf(" SDMMC\n");
	else if (OmPin == BOOT_EMMC)
		printf(" EMMC\n");
	else if (OmPin == BOOT_EMMC_4_4)
		printf(" EMMC\n");
	else
		printf(" Please check OM_pin\n");
}

int board_init(void)
{
	int ret = 0;

	init_nr_dram_banks();

	display_bl1_version();

	display_pmic_info();

	display_boot_device_info();

	gd->bd->bi_boot_params = (PHYS_SDRAM_1 + 0x100UL);
	/* Initializing Audio */
	ret = sound_init();

	return ret;
}

int dram_init(void)
{
	gd->ram_size	= get_ram_size((long *)PHYS_SDRAM_1, PHYS_SDRAM_1_SIZE)
			+ get_ram_size((long *)PHYS_SDRAM_2, PHYS_SDRAM_2_SIZE)
			+ get_ram_size((long *)PHYS_SDRAM_3, PHYS_SDRAM_3_SIZE)
			+ get_ram_size((long *)PHYS_SDRAM_4, PHYS_SDRAM_4_SIZE)
			+ get_ram_size((long *)PHYS_SDRAM_5, PHYS_SDRAM_7_SIZE)
			+ get_ram_size((long *)PHYS_SDRAM_6, PHYS_SDRAM_7_SIZE)
			+ get_ram_size((long *)PHYS_SDRAM_7, PHYS_SDRAM_7_SIZE);

	if (nr_dram_banks == 8)
		gd->ram_size += get_ram_size((long *)PHYS_SDRAM_8, PHYS_SDRAM_8_END_SIZE);
	else {
		gd->ram_size += get_ram_size((long *)PHYS_SDRAM_8, PHYS_SDRAM_8_SIZE)
			+ get_ram_size((long *)PHYS_SDRAM_9, PHYS_SDRAM_9_SIZE)
			+ get_ram_size((long *)PHYS_SDRAM_10, PHYS_SDRAM_10_SIZE)
			+ get_ram_size((long *)PHYS_SDRAM_11, PHYS_SDRAM_11_SIZE)
			+ get_ram_size((long *)PHYS_SDRAM_12, PHYS_SDRAM_12_SIZE);
	}

	return 0;
}

void dram_init_banksize(void)
{
	gd->bd->bi_dram[0].start = PHYS_SDRAM_1;
	gd->bd->bi_dram[0].size = get_ram_size((long *)PHYS_SDRAM_1,
							PHYS_SDRAM_1_SIZE);
	gd->bd->bi_dram[1].start = PHYS_SDRAM_2;
	gd->bd->bi_dram[1].size = get_ram_size((long *)PHYS_SDRAM_2,
							PHYS_SDRAM_2_SIZE);
	gd->bd->bi_dram[2].start = PHYS_SDRAM_3;
	gd->bd->bi_dram[2].size = get_ram_size((long *)PHYS_SDRAM_3,
							PHYS_SDRAM_3_SIZE);
	gd->bd->bi_dram[3].start = PHYS_SDRAM_4;
	gd->bd->bi_dram[3].size = get_ram_size((long *)PHYS_SDRAM_4,
							PHYS_SDRAM_4_SIZE);
	gd->bd->bi_dram[4].start = PHYS_SDRAM_5;
	gd->bd->bi_dram[4].size = get_ram_size((long *)PHYS_SDRAM_5,
							PHYS_SDRAM_5_SIZE);
	gd->bd->bi_dram[5].start = PHYS_SDRAM_6;
	gd->bd->bi_dram[5].size = get_ram_size((long *)PHYS_SDRAM_6,
							PHYS_SDRAM_6_SIZE);
	gd->bd->bi_dram[6].start = PHYS_SDRAM_7;
	gd->bd->bi_dram[6].size = get_ram_size((long *)PHYS_SDRAM_7,
							PHYS_SDRAM_7_SIZE);

	if (nr_dram_banks == 8) {
		gd->bd->bi_dram[7].start = PHYS_SDRAM_8;
		gd->bd->bi_dram[7].size = get_ram_size((long *)PHYS_SDRAM_8,
							PHYS_SDRAM_8_END_SIZE);
	} else if (nr_dram_banks == 12) {
		gd->bd->bi_dram[7].start = PHYS_SDRAM_8;
		gd->bd->bi_dram[7].size = get_ram_size((long *)PHYS_SDRAM_8,
							PHYS_SDRAM_8_SIZE);
		gd->bd->bi_dram[8].start = PHYS_SDRAM_9;
		gd->bd->bi_dram[8].size = get_ram_size((long *)PHYS_SDRAM_9,
							PHYS_SDRAM_9_SIZE);
		gd->bd->bi_dram[9].start = PHYS_SDRAM_10;
		gd->bd->bi_dram[9].size = get_ram_size((long *)PHYS_SDRAM_10,
							PHYS_SDRAM_10_SIZE);
		gd->bd->bi_dram[10].start = PHYS_SDRAM_11;
		gd->bd->bi_dram[10].size = get_ram_size((long *)PHYS_SDRAM_11,
							PHYS_SDRAM_11_SIZE);
		gd->bd->bi_dram[11].start = PHYS_SDRAM_12;
		gd->bd->bi_dram[11].size = get_ram_size((long *)PHYS_SDRAM_12,
							PHYS_SDRAM_12_SIZE);
	}
}

#ifdef CONFIG_DISPLAY_BOARDINFO
int checkboard(void)
{
	board_id_t id = get_board_info();
	uint32_t rev = id.revision & 0x7;
	printf("\nBoard: KMHUB5422\n");
	printf("Variant: %s - revision 0x%02x\n",
			(id.is_dev_board) ? "Dev board" : "Form Factor",
			rev);

	return 0;
}
#endif

#ifdef CONFIG_GENERIC_MMC
int board_mmc_init(bd_t *bis)
{
	struct exynos5_power *pmu = (struct exynos5_power *)EXYNOS5_POWER_BASE;
	int err, OmPin;

	OmPin = readl(&pmu->inform3);

	// Force eMMC external LDO enable

	// GPC0[2] to output mode - driving high
	rGPC0CON = (rGPC0CON & ~(0xf << 8)) | (0x1 << 8);
	rGPC0DAT |= (1 << 2);

	// 10ms power-on time (a total guess)
	sdelay(1000000);

	// Drive RESET_L high
	rGPD1CON = (rGPD1CON & ~0xf) | 0x1;
	rGPD1DAT |= 1;
	// Disable pull-down
	rGPD1PUD = (rGPD1PUD & ~0x3);

	// 10ms more POR time (another guess)
	sdelay(1000000);

	err = exynos_pinmux_config(PERIPH_ID_SDMMC0, PINMUX_FLAG_8BIT_MODE);
	if (err) {
		debug("MSHC0 not configured\n");
		return err;
	}

	switch (OmPin) {
	case BOOT_EMMC_4_4:
	default:
#if defined(USE_MMC0)
		set_mmc_clk(PERIPH_ID_SDMMC0, 1);

		err = s5p_mmc_init(PERIPH_ID_SDMMC0, 8);
#endif
		break;
	}

	return err;
}
#endif

static int board_uart_init(void)
{
	int err;

	err = exynos_pinmux_config(PERIPH_ID_UART3, PINMUX_FLAG_NONE);
	if (err) {
		debug("UART3 not configured\n");
		return err;
	}

	err = exynos_pinmux_config(PERIPH_ID_UART2, PINMUX_FLAG_NONE);
	if (err) {
		debug("UART2(bt-uart) could not configured\n");
		return err;
	}

	return 0;
}


static void setup_pins(gpio_bank_list_t *list, int num) {

	int idx;
	uint8_t pin_idx = 0;

	for(idx = 0; idx < num; idx++) {
		for(pin_idx = 0; pin_idx < 8 && list[idx].pin_cfgs[pin_idx].pin != ((uint8_t)-1U); pin_idx++) {
			gpio_pin_cfg_t *p_cfg = &(list[idx].pin_cfgs[pin_idx]);
			struct s5p_gpio_bank *bank = list[idx].bank;

			if(p_cfg->cfg != GPIO_DEFAULT) {
				s5p_gpio_cfg_pin(bank, p_cfg->pin, p_cfg->cfg);
			}

			if(p_cfg->pud != GPIO_DEFAULT) {
				s5p_gpio_set_pull(bank, p_cfg->pin, p_cfg->pud);
			}

			if(p_cfg->drv != GPIO_DEFAULT) {
				s5p_gpio_set_drv(bank, p_cfg->pin, p_cfg->drv);
			}

			if(p_cfg->rate != GPIO_DEFAULT) {
				s5p_gpio_set_rate(bank, p_cfg->pin, p_cfg->rate);
			}
			
			if(p_cfg->pdn_con != GPIO_DEFAULT) {
				s5p_gpio_pdn_cfg_pin(bank, p_cfg->pin, p_cfg->pdn_con);
			}

			if(p_cfg->pdn_pull != GPIO_DEFAULT) {
				s5p_gpio_set_pdn_pull(bank, p_cfg->pin, p_cfg->pdn_pull);
			}
		}
	}
}

static void setup_default_pins(void) {
	struct exynos5422_gpio_part1 *gp1 = (struct exynos5422_gpio_part1 *)samsung_get_base_gpio_part1();
	struct exynos5420_gpio_part2 *gp2 = (struct exynos5420_gpio_part2 *)samsung_get_base_gpio_part2();
	struct exynos5420_gpio_part3 *gp3 = (struct exynos5420_gpio_part3 *)samsung_get_base_gpio_part3();
	struct exynos5420_gpio_part4 *gp4 = (struct exynos5420_gpio_part4 *)samsung_get_base_gpio_part4();

	gpio_bank_list_t default_pin_cfgs[] = {
		{
			.bank = &gp1->y7,
			.pin_cfgs = {
				{7, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp1->etc6,
			.pin_cfgs = {
				{6, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp1->etc5,
			.pin_cfgs = {
				{0, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{1, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{4, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{5, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp1->x0,
			.pin_cfgs = {
				{0, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{1, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{5, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{7, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp1->x1,
			.pin_cfgs = {
				{0, GPIO_INPUT, GPIO_PULL_UP, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{1, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{3, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{5, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{7, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp1->x2,
			.pin_cfgs = {
				{0, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{1, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp2->c0,
			.pin_cfgs = {
				{3, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{4, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{5, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{6, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{7, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp2->c1,
			.pin_cfgs = {
				{0, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{1, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{3, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{4, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{5, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{6, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp2->c2,
			.pin_cfgs = {
				{1, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{3, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{4, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_DOWN},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp2->c3,
			.pin_cfgs = {
				{0, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{1, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{2, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{3, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp2->c4,
			.pin_cfgs = {
				{0, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_DOWN},
				{1, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp3->g0,
			.pin_cfgs = {
				{0, GPIO_INPUT, GPIO_PULL_DOWN, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{1, GPIO_INPUT, GPIO_PULL_UP, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				{2, GPIO_INPUT, GPIO_PULL_UP, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				{3, GPIO_INPUT, GPIO_PULL_UP, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				{4, GPIO_INPUT, GPIO_PULL_UP, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				{5, GPIO_INPUT, GPIO_PULL_DOWN, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{6, GPIO_INPUT, GPIO_PULL_DOWN, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{7, GPIO_INPUT, GPIO_PULL_DOWN, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
			},
		},
		{
			.bank = &gp3->g1,
			.pin_cfgs = {
				{5, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_DOWN},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp3->g2,
			.pin_cfgs = {
				{0, GPIO_INPUT, GPIO_PULL_DOWN, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{1, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp4->a0,
			.pin_cfgs = {
				{0, GPIO_INPUT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_DEFAULT},
				{1, GPIO_INPUT, GPIO_PULL_DOWN, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp4->a1,
			.pin_cfgs = {
				{0, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				{1, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				{2, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				{3, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				{4, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_DEFAULT},
				{5, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp4->a2,
			.pin_cfgs = {
				{0, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_DEFAULT},
				{1, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_DEFAULT},
				{4, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{6, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				{7, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp4->b3,
			.pin_cfgs = {
				{6, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_DEFAULT},
				{7, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp4->b4,
			.pin_cfgs = {
				{0, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_DEFAULT},
				{1, GPIO_DEFAULT, GPIO_PULL_NONE, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_DEFAULT},
				TERMINATE_PINLIST,
			},
		},
		{
			.bank = &gp4->h0,
			.pin_cfgs = {
				{0, GPIO_INPUT, GPIO_PULL_UP, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				{1, GPIO_INPUT, GPIO_PULL_DOWN, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_DOWN},
				{2, GPIO_INPUT, GPIO_PULL_DOWN, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_DOWN},
				{3, GPIO_INPUT, GPIO_PULL_UP, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				{4, GPIO_INPUT, GPIO_PULL_UP, GPIO_DEFAULT, GPIO_DEFAULT, GPIO_PDN_PREV_STATE, GPIO_PULL_UP},
				TERMINATE_PINLIST,
			},
		},
	};

	int num_default_cfgs = sizeof(default_pin_cfgs)/ sizeof(default_pin_cfgs[0]);

	setup_pins(default_pin_cfgs, num_default_cfgs);
}

#ifdef CONFIG_BOARD_EARLY_INIT_F
int board_early_init_f(void)
{
	init_nr_dram_banks();

	timer_init();

	// FORCE_DFU Sensing

	// GPX1[0] to input (TP2035 on Dev, FORCE_DFU_L on FF)
	rGPX1CON = rGPX1CON & ~(0xF);
	// GPX1[0] pull-up enabled
	rGPX1PUD |= 0x3;

	// Board ID / Rev Sensing
	/*
	 * Configure GPG0_1-4 as inputs - all GPC pins are either board ID or NC
	 */
	rGPG0CON = 0;

	/*
	 * Enable pull-downs on all BOARD_ID and BOARD_REV[0:2] - all GPG pins
	 * are NC otherwise, so pull-downs on all pins is okay.
	 */
	rGPG0PUD = 0x5555;

	int ret = board_uart_init();

	(void)get_board_info();

	return ret;
}
#endif

#if 0
unsigned int    filecheck(const unsigned char *filename)
{
    unsigned char   cmd[128];
    unsigned long   filesize = 0;

    memset(cmd, 0x00, sizeof(cmd));

    // file check & update
	setenv("filesize", "0");

    sprintf(cmd, "fatload mmc 0:1 40008000 update/%s", filename);

    run_command(cmd, 0);

    if((filesize = getenv_ulong("filesize", 16, 0)))    return  1;  // file read success

    printf("ERROR! update/%s File Not Found!! filesize = 0\n", filename);

    return  0;  // file read fail
}

void update_image       (const unsigned char *partition)
{
    unsigned char   cmd[128];

    memset(cmd, 0x00, sizeof(cmd));
    sprintf(cmd, "fastboot flash %s 40008000 0", partition);
    run_command(cmd, 0);
}

void update_raw_image   (const unsigned char *partition)
{
	struct exynos5_power *pmu = (struct exynos5_power *)EXYNOS5_POWER_BASE;
	int OmPin;

	OmPin = readl(&pmu->inform3);

    if(!strncmp(partition, "kernel", sizeof("kernel")))     run_command("movi write kernel 0 40008000", 0);
    else    {
    	if (OmPin == BOOT_EMMC_4_4 || OmPin == BOOT_EMMC)   {
        	run_command("emmc open 0", 0);
            if(!strncmp(partition, "bootloader",    sizeof("bootloader")))  run_command("movi w z u 0 40008000", 0);
            if(!strncmp(partition, "bl1",           sizeof("bl1")))         run_command("movi w z f 0 40008000", 0);
            if(!strncmp(partition, "bl2",           sizeof("bl2")))         run_command("movi w z b 0 40008000", 0);
            if(!strncmp(partition, "tzsw",          sizeof("tzsw")))        run_command("movi w z t 0 40008000", 0);
            run_command("emmc close 0", 0);
    	}
    	else    {
            if(!strncmp(partition, "bootloader",    sizeof("bootloader")))  run_command("movi w u 0 40008000", 0);
            if(!strncmp(partition, "bl1",           sizeof("bl1")))         run_command("movi w f 0 40008000", 0);
            if(!strncmp(partition, "bl2",           sizeof("bl2")))         run_command("movi w b 0 40008000", 0);
            if(!strncmp(partition, "tzsw",          sizeof("tzsw")))        run_command("movi w t 0 40008000", 0);
    	}
    }
}

#define OPTION_ERASE_USERDATA   0x01
#define OPTION_ERASE_FAT        0x02
#define OPTION_ERASE_ENV        0x04
#define OPTION_UPDATE_UBOOT     0x08

int board_late_init(void)
{
	struct exynos5_power *pmu = (struct exynos5_power *)EXYNOS5_POWER_BASE;

    LED_GREEN(OFF);          LED_BLUE(ON);          LED_RED(OFF);

    if (readl(&pmu->sysip_dat0) & 0xFFFF)     {
        unsigned char   cmd[32];
        unsigned int    reboot_cmd;
        unsigned int    option;
        unsigned long   filesize;

        #define INFORM0_REG 0x10040800
        #define INFORM2_REG 0x10040808
        #define INFORM5_REG 0x10040814  /* SYSIP_DAT1 */
        #define INFORM6_REG 0x10040818  /* SYSIP_DAT2 */
        #define INFORM7_REG 0x1004081C  /* SYSIP_DAT3 */

        reboot_cmd = readl(&pmu->sysip_dat0);
        writel(0x00, &pmu->sysip_dat0);

        option = readl(INFORM0_REG);
        writel(0x00, INFORM0_REG);

        memset(cmd, 0x00, sizeof(cmd));

        switch(reboot_cmd  & 0xFFFF) {
            case    REBOOT_FASTBOOT:
                sprintf(cmd, "%s", "fastboot");
                run_command(cmd, 0);
                break;
            case    REBOOT_UPDATE:
#if defined(CONFIG_LED_CONTROL)
                LED_BLUE(ON);
#endif
                if(filecheck("system.img"))     update_image("system");
                if(filecheck("cache.img"))      update_image("cache");

                if(option & OPTION_ERASE_USERDATA)  {
                    if(filecheck("userdata.img"))   update_image("userdata");
                }

                if      (filecheck("zImage"))       update_raw_image("kernel");
                else if (filecheck("zImage-dtb"))   update_raw_image("kernel");

                if(option & OPTION_UPDATE_UBOOT)    {
                    if(filecheck("u-boot.bin")) update_raw_image("bootloader");
                    if(filecheck("bl1.bin"))    update_raw_image("bl1");
                    if(filecheck("bl2.bin"))    update_raw_image("bl2");
                    if(filecheck("tzsw.bin"))   update_raw_image("tzsw");
                }

                if(option & OPTION_ERASE_ENV)
                    run_command("mmc write 0 40008000 0x4CF 0x20", 0);

                if(option & OPTION_ERASE_FAT)
                    run_command("fatformat mmc 0:1", 0);

#if defined(CONFIG_LED_CONTROL)
                LED_BLUE(OFF);
#endif
                break;
            default :
                break;
        }

        // Restart system
        memset(cmd, 0x00, sizeof(cmd));
        sprintf(cmd, "%s", "reset");
        run_command(cmd, 0);
    }

	return 0;
}

#endif

int board_late_init(void)
{
	struct exynos5_power *pmu = (struct exynos5_power *)EXYNOS5_POWER_BASE;
	unsigned int rst_stat = 0;

	/* Check if we're on a FF board */
	board_id_t id = get_board_info();

	uint8_t is_force_dfu = 0;

	if(!id.is_dev_board) {
		/* FF board - check for FORCE_DFU being LOW (active low) */

#if !defined(CONFIG_BOARD_EARLY_INIT_F)
		// If this wasn't already done
		// GPX1[0] to input,
		rGPX1CON = rGPX1CON & ~(0xF);
		// GPX1[0] pull-up enabled
		rGPX1PUD |= 0x3;
#endif
		/* ~ 10ms RC Settle time.. */
		sdelay(1000000); // Measured on scope - 100000 sdelay is ~1.1ms

		if(!(rGPX1DAT & 0x1)) {
			is_force_dfu = 1;
		}
	}

	if(is_force_dfu) {
		printf("FORCE_DFU_L is asserted..\n");

		// FORCE_DFU is asserted - set bootcmd to our DFU mode command
		setenv("bootcmd", CONFIG_BOOTCOMMAND_FF_DFU);
		run_command("run bootcmd", 0);
	} else {
		printf("FORCE_DFU_L is NOT asserted..\n");
		setenv("bootcmd", CONFIG_BOOTCOMMAND);
	}

#if 0
	/* eMMC NUKE mode */
	setenv("bootcmd", CONFIG_EMMC_NUKE_BOOTCOMMAND);
#endif

	rst_stat = readl(&pmu->rst_stat);
	printf("rst_stat : 0x%x\n", rst_stat);
#ifdef CONFIG_RAMDUMP_MODE
	/* check reset status for ramdump */
	if ((rst_stat & (WRESET | SYS_WDTRESET))
		|| (readl(&pmu->sysip_dat0) == CONFIG_RAMDUMP_MODE)
		|| (readl(CONFIG_RAMDUMP_SCRATCH) == CONFIG_RAMDUMP_MODE)) {
		/* run fastboot */
		run_command("fastboot", 0);
	}
#endif
#ifdef CONFIG_RECOVERY_MODE
	u32 second_boot_info = readl(CONFIG_SECONDARY_BOOT_INFORM_BASE);

	if (second_boot_info == 1) {
		printf("###Recovery Mode###\n");
		writel(0x0, CONFIG_SECONDARY_BOOT_INFORM_BASE);
		setenv("bootcmd", CONFIG_BOOTCOMMAND_FF_DFU);
	}
#endif

	if ((readl(&pmu->sysip_dat0)) == CONFIG_FACTORY_RESET_MODE) {
		writel(0x0, &pmu->sysip_dat0);
		setenv("bootcmd", CONFIG_FACTORY_RESET_BOOTCOMMAND);
	}

	// Kamama: RDR-1183 - Properly configure GPIO Pullups/downs
	setup_default_pins();

	return 0;
}

unsigned int get_board_rev(void)
{
	board_id_t id = get_board_info();

	return id.revision;
}
